In order to demonstrate the functionality of PROLEAD, we provide some selected examples in the <a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples">examples</a> folder.
More information regarding the interpretation of results is given under <a href="https://github.com/ChairImpSec/PROLEAD/wiki/Results">Results</a>.

# Hardware Examples
Each example is structured as follows:

1. The `rtl` folder contains the rtl code in VHDL or Verilog.
2. The `gate` folder contains the resulting gate level netlist which is given as an input to PROLEAD. 
3. The `results` directory which is further divided into `normal_mode` and `compact_mode` directories contains the used `config.set` file which encompasses all evaluation settings. In addition, we created an execution script `run.sh` for each test case. To execute the script, just navigate into the respective folder and run `./run.sh`. Moreover, `results` contains all reports generated by PROLEAD. The main report is given as `Report.dat` and builds the starting point for interpreting the results.

We summarize the evaluation results in the following table. We remark that the configuration can be slightly different compared to the paper. Note that different configurations may lead to different runtimes and memory requirements. Moreover, runtimes can be different when using another machine. All examples were performed in a Ubuntu 20.04 subsystem on a Windows 10 server with 544GB RAM and by using at most 48 cores.

 | Design | Reference | Order | Secure | RAM (Compact) | Time (Compact) | RAM (Normal) | Time (Normal) |
 |:------:|:---------:|:-----:|:------:|:-------------:|:--------------:|:------------:|:-------------:|
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/PRESENT_Sbox_TI_d1_Uniform/rtl">TI, PRESENT S-box, Uniform</a> | <a href="https://link.springer.com/article/10.1007/s00145-010-9086-6">[PMK+11]</a> | 1 | :heavy_check_mark:| 3.8GB | 30.6sec | 4.0GB | 0.1sec |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/PRESENT_Sbox_TI_d1_Not_Uniform/rtl">TI, PRESENT S-box, Not uniform</a> | <a href="https://link.springer.com/chapter/10.1007/978-3-319-70694-8_26">[EGMP17]</a> | 1 | :x: | 3.5GB | 0.1sec | 3.5GB | 0.1sec |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Sbox_TI_d1/rtl">TI, AES S-Box</a> | <a href="https://link.springer.com/chapter/10.1007/978-3-642-20465-4_6">[MPL+11]</a> | 1 | :heavy_check_mark:| 6.3GB | 6.6min | 37.3GB | 3.6min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Sbox_DOM_d1/rtl">DOM, AES S-Box</a> | <a href="https://dl.acm.org/doi/10.1145/2996366.2996426">[GMK16]</a> | 1 | :heavy_check_mark:| 6.3GB | 6.2min | 12.4GB | 2.1min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Sbox_DOM_d2/rtl">DOM, AES S-Box</a> | <a href="https://dl.acm.org/doi/10.1145/2996366.2996426">[GMK16]</a> | 2 | :heavy_check_mark:| 8.4GB | 3.9h | 492.9GB | 3.7h |
|<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Sbox_CMS_d1/rtl">CMS, AES S-Box</a> | <a href="https://link.springer.com/chapter/10.1007/978-3-662-53140-2_10">[CRB+16]</a> | 1 | :heavy_check_mark:| 6.3GB | 8.7min | 38.4GB | 4.1min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Sbox_CMS_d2/rtl">CMS, AES S-Box</a> | <a href="https://link.springer.com/chapter/10.1007/978-3-662-53140-2_10">[CRB+16]</a> | 2 | :heavy_check_mark:| 8.8GB | 4.8h | 504.3GB | 5.4h |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/PRESENT_Enc_TI_d1_Uniform/rtl">TI, nibble-serial PRESENT-80</a> | <a href="https://link.springer.com/article/10.1007/s00145-010-9086-6">[PMK+11]</a> | 1 | :heavy_check_mark:| 36.9GB | 49.4min | 5.6GB | 2.3min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/PRESENT_Enc_TI_d1_Uniform/rtl">TI, nibble-serial PRESENT-80</a> | <a href="https://link.springer.com/chapter/10.1007/978-3-319-70694-8_26">[EGMP17]</a> | 1 | :x:| 3.7GB | 1.2sec | 3.9GB | 4.9sec |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Enc_TI_d1/rtl">TI, byte-serial AES-128</a> | <a href="https://link.springer.com/chapter/10.1007/978-3-642-20465-4_6">[MPL+11]</a> | 1 | :heavy_check_mark:| 78.4GB | 3.6h | 298.7GB | 1.2h |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Enc_DOM_d1/rtl">DOM, byte-serial AES-128</a> | <a href="https://dl.acm.org/doi/10.1145/2996366.2996426">[GMK16]</a> | 1 | :heavy_check_mark:| 47.7GB | 1.8h | 143.1GB | 28.4min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_Enc_CMS_d1/rtl">CMS, byte-serial AES-128</a> | <a href="https://link.springer.com/chapter/10.1007/978-3-662-53140-2_10">[CRB+16]</a> | 1 | :heavy_check_mark:| 47.7GB | 1.8h | 263.7GB | 36.0min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/MIDORI_NullFresh_d1/rtl">NullFresh, Midori-64</a> | <a href="https://tches.iacr.org/index.php/TCHES/article/view/8736">[SM21a]</a> | 1 | :heavy_check_mark:| 17.0GB | 20.2min | 127.9GB | 13.3min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/PRESENT_NullFresh_d1/rtl">NullFresh, PRESENT-80</a> | <a href="https://tches.iacr.org/index.php/TCHES/article/view/8736">[SM21a]</a> | 1 | :x: | 7.0GB | 16.6sec | 3.8GB | 5.4sec |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/PRINCE_NullFresh_d1/rtl">NullFresh, PRINCE</a> | <a href="https://tches.iacr.org/index.php/TCHES/article/view/8736">[SM21a]</a> | 1 | :x: | 18.6GB | 2.3min | 14.8GB | 4.1min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/AES_NullFresh_d1/rtl">NullFresh, AES-128</a> | <a href="https://tches.iacr.org/index.php/TCHES/article/view/8736">[SM21a]</a> | 1 | :x: | 5.5GB | 10.5sec | 78.8GB | 21.0min |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/KECCAK_Low_Latency_d1/rtl">Low-Latency Keccak-f[25]</a> | <a href="https://tches.iacr.org/index.php/TCHES/article/view/9070/8654">[ZSS+21]</a> | 1 | :x: | 6.5GB | 49.8sec | 5.6GB | 22.2sec |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/KECCAK_Low_Latency_d2/rtl">Low-Latency Keccak-f[25]</a> | <a href="https://tches.iacr.org/index.php/TCHES/article/view/9070/8654">[ZSS+21]</a> | 2 | :x: | 17.5GB | 3.1day | 531.9GB | 1.6day |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/LowRand_LED_3Stage_d2/rtl">Low-rand, LED-128, 3-stage</a> | <a href="https://tches.iacr.org/index.php/TCHES/article/view/9312">[BDMS22]</a> | 2 | :x: | 14.7GB | 30.4min | 201.5GB | 2.6h |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/GHPC_Gadgets_d1_wo_Optional_Regs/rtl">GHPC gadgets, without optional regs</a> | <a href="https://eprint.iacr.org/2021/247.pdf">[KSM22]</a> | 1 | :x: | 3.5GB | 0.1sec | 3.5GB | 0.1sec |
 |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Hardware/GHPC_Gadgets_d1_w_Optional_Regs/rtl">GHPC gadgets, with optional regs</a> | <a href="https://eprint.iacr.org/2021/247.pdf">[KSM22]</a> | 1 | :heavy_check_mark:| 4.1GB | 49.6sec | 3.5GB | 0.1sec |

# Software Examples
Each example is structured as follows:

1. The `source` folder contains the C/Assembly code and the used linker file.
2. The `binary` folder contains the resulting binary, disassembled file and map file which is given as an input to PROLEAD. 
3. The `results` directory which is further divided into `normal_mode` and `compact_mode` directories contains the used `config.set` file which encompasses all evaluation settings. In addition, we created an execution script `run.sh` for each test case. To execute the script, just navigate into the respective folder, make the script executable and run `./run.sh`. Moreover, `results` contains all reports generated by PROLEAD. The main report is given as `Report.dat` and builds the starting point for interpreting the results.

We summarize the evaluation results in the following table. The config.set in each directory describes the settings used for the evaluation of the testcases in our paper. Furthermore in our paper we ceased an evaluation if we identify significant leakage that increases after every simulation. We have set the border to abort the evaluation when we crossed a *g*-value of 100. The table entries from `ISW` to `PFE` describe the security definitions that were defined during the construction of the CPU-independent leakage model in the paper. We remark that the configuration can be slightly different compared to the paper. Note that different configurations may lead to different runtimes and memory requirements. Moreover, runtimes can be different when using another machine. All examples were performed in a Ubuntu 20.04 subsystem on a Windows 10 server with 544GB RAM and by using at most 30 cores. We compiled the source files with the 9-2019-q4-major arm-none-eabi-gcc version.


 | Design | Reference | Order | ISW | MOE | MRE | BIL | NLE | PRO | PFE | RAM (Normal) | Time (Normal) |
 |:------:|:---------:|:-----:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:------------:|:-------------:|
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_AES_rp">AES, RP</a> | <a href="https://doi.org/10.1007/978-3-642-15031-9\_28">[RP10]</a> | 1 | :heavy_check_mark: | :x: | :x: | :x: | :x: | :x: | :x: | 5.5GB | 14sec |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_AES_htable">AES, Htable</a> | <a href="https://doi.org/10.1007/978-3-642-55220-5\_25">[Cor14]</a> | 1 | :heavy_check_mark: | :x: | :x: | :x: | :x: | :x: | :x: | 9.5GB | 1min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_AES_ipm">AES, IPM</a> | <a href="https://doi.org/10.29007/fv2n">[CCG+21]</a> | 1 | :x: | :x: | :x: | :x: | :x: | :x: | :x: | 13GB | 35min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_AES_flr_mprgmat">AES, FLR</a> | <a href="https://doi.org/10.1007/978-3-030-45727-3\_12">[CGZ20]</a> | 1 | :x: | :x: | :x: | :x: | :x: | :x: | :x: | 5.8GB | 22sec |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_AES_anssi">AES, ANSSI</a> | <a href="http://eprint.iacr.org/2010/523">[FMPR10]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | 21GB | 17min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_SECADD">SecADD</a> | <a href="https://doi.org/10.46586/tches.v2022.i4.553-588">[BC22]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | 18.3GB | 27min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_B2A_MODP">SecB2AModp</a> | <a href="https://doi.org/10.46586/tches.v2022.i4.553-588">[BC22]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :x: | :x: | :x: | 123GB | 21h |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_A2B_MODP">SecA2BModp</a> | <a href="https://doi.org/10.46586/tches.v2022.i4.553-588">[BC22]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :x: | :x: | :x: | 45GB | 10min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_PRESENT_TI">PRESENT</a> | <a href="https://doi.org/10.46586/tches.v2023.i2.155-179">[GD22]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :x: | :x: | :x: | :x: | 14GB | 4min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_keccak_f1600_TI">Keccak-f1600</a> | <a href="https://doi.org/10.46586/tches.v2023.i2.155-179">[GD22]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :x: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | 95GB | 29min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_keccak_f800_TI">Keccak-f800</a> | <a href="https://doi.org/10.46586/tches.v2023.i2.155-179">[GD22]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :x: | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | 45GB | 14min |
  |<a href="https://github.com/ChairImpSec/PROLEAD/tree/main/examples/Software/SW_quadratic_classes_concat">Quadratic Class</a> | <a href="https://doi.org/10.46586/tches.v2023.i2.155-179">[GD22]</a> | 1 | :heavy_check_mark: | :heavy_check_mark: | :heavy_check_mark: | :x: | :heavy_check_mark: | :heavy_check_mark: | :x: | 8GB | 3min |
