Welcome to the PROLEAD installation guide! This comprehensive guide will walk you through the step-by-step process of installing PROLEAD on your system.

# PROLEAD
We want to emphasize that the installation of PROLEAD_SW alone is not supported. To successfully install PROLEAD_SW, you must perform a full installation of PROLEAD. In order to install PROLEAD for leakage detection, we refer to the following guide.

## Requirements
We tested the following installation guide on an Ubuntu-22.04 subsystem (wsl 2) with the following requirements:

`git 2.25.1` `GNU Make 4.2.1` `g++ 11.3.0` `boost 1.71` `python 3` `arm-none-eabi-gcc` `pkg-config`

If your system is missing any of the required tools, we highly recommend following the instructions below to install them. These instructions will guide you through the installation process of the missing requirements, ensuring that your system meets all the necessary prerequisites.

| Instruction | Usage |
|:-----------:|:-----:|
| `sudo apt-get install git` | Installs git |
| `sudo apt-get install make` | Installs make |
| `sudo apt-get install g++` | Installs g++ |
| `sudo apt-get install libboost-all-dev` | Installs boost |
| `sudo apt-get install -y gcc-arm-none-eabi` | Installs the arm gcc toolchain |
| `sudo apt-get install pkg-config` | Installs the pkg tool |

## Installation
Please follow the instructions below in order to install PROLEAD:
1. `git clone https://github.com/ChairImpSec/PROLEAD` to clone the git repository
2. `cd PROLEAD` to change to the "PROLEAD" folder
3. `make release` to build PROLEAD

# Extensions
To utilize Yosys for synthesis, it is necessary to install a separate toolchain.

## Icarus Verilog
<a href="https://github.com/steveicarus/iverilog">Icarus Verilog</a> is an open-source compiler implementation for Verilog. For more information regarding the installation, we refer to the original <a href="https://iverilog.fandom.com/wiki/Installation_Guide">installation guide</a>.

### Requirements
We tested the following installation guide on an Ubuntu-22.04 subsystem (wsl 2) with the following requirements:

`GNU gperf 3.1` `autoconf (GNU Autoconf) 2.71` 

If your system is missing any of the required tools, we highly recommend following the instructions below to install them. These instructions will guide you through the installation process of the missing requirements, ensuring that your system meets all the necessary prerequisites.

| Instruction | Usage |
|:-----------:|:-----:|
| `sudo apt-get install gperf` | Installs gperf |
| `sudo apt-get install autoconf` | Installs autoconf |

### Installation
Please follow the instructions below in order to install <a href="https://github.com/steveicarus/iverilog">Icarus Verilog</a>:
1. `git clone https://github.com/steveicarus/iverilog.git` to clone the git repository
2. `cd iverilog` to change to the "iverilog" folder
3. `sh autoconf.sh` to build the configuration files
4. `./configure` to run the configuration script  
5. `make` to build <a href="https://github.com/steveicarus/iverilog">Icarus Verilog</a>
6. `make install` to install <a href="https://github.com/steveicarus/iverilog">Icarus Verilog</a>

## Yosys
In order to install <a href="https://github.com/YosysHQ/yosys">Yosys</a> for synthesis, we refer to the following guide.

### Requirements
We tested the following installation guide on an Ubuntu-22.04 subsystem (wsl 2) with the following requirements:
`clang 14.0.0-1ubuntu1` `bison (GNU Bison) 3.8.2` `flex 2.6.4` `GNU Awk 5.1.0` `Python 3.10.6`

If your system is missing any of the required tools, we highly recommend following the instructions below to install them. These instructions will guide you through the installation process of the missing requirements, ensuring that your system meets all the necessary prerequisites.

| Instruction | Usage |
|:-----------:|:-----:|
| `sudo apt-get install clang` | Installs clang
| `sudo apt-get install bison` | Installs bison
| `sudo apt-get install flex` | Installs flex
| `sudo apt-get install gawk` | Installs gawk
| `sudo apt-get install python3` | Installs python
| `sudo apt-get install build-essential` |
| `sudo apt-get install libreadline-dev` |
| `sudo apt-get install tcl-dev` |
| `sudo apt-get install libffi-dev` |
| `sudo apt-get install graphviz` |
| `sudo apt-get install xdot` |
| `sudo apt-get install pkg-config` |
| `sudo apt-get install zlib1g-dev` |

### Installation
Please follow the instructions below in order to install <a href="https://github.com/YosysHQ/yosys">Yosys</a>:
1. `git clone https://github.com/YosysHQ/yosys` to clone the git repository
2. `cd yosys` to change to the "yosys" directory
3. `make` to build <a href="https://github.com/YosysHQ/yosys">Yosys</a>
4. `make install` to install <a href="https://github.com/YosysHQ/yosys">Yosys</a>
5. `make test` to execute tests

## GHDL
The public version of <a href="https://github.com/YosysHQ/yosys">Yosys</a> can only synthesize designs written in Verilog. To synthesize projects written in VHDL, <a href="https://github.com/ghdl/ghdl">GHDL</a> is required. 

### Requirements
We tested the following installation guide on an Ubuntu-22.04 subsystem (wsl 2) with the following requirements:

`gcc 11.3.0` `GNAT 10.4.0` 

If your system is missing any of the required tools, we highly recommend following the instructions below to install them. These instructions will guide you through the installation process of the missing requirements, ensuring that your system meets all the necessary prerequisites.

| Instruction | Usage |
|:-----------:|:-----:|
| `sudo apt-get install gcc` | Installs gcc
| `sudo apt-get install gnat` | Installs gnat

### Installation
Please follow the instructions below in order to install <a href="https://github.com/ghdl/ghdl">GHDL</a> or visit the <a href="https://ghdl.github.io/ghdl/development/building/index.html">GHDL Installation Guide</a> 

1. `git clone https://github.com/ghdl/ghdl` to clone the git repository
2. `cd ghdl` to change to the "ghdl" directory
3. `mkdir build` to create the "build" directory
4. `cd build` to change to the "build" directory
5. `../configure --prefix=PREFIX` to configure <a href="https://github.com/ghdl/ghdl">GHDL</a>
6. `make` to build <a href="https://github.com/ghdl/ghdl">GHDL</a>
7. `make install` to install <a href="https://github.com/ghdl/ghdl">GHDL</a>

## ghdl-yosys-plugin
This <a href="https://github.com/ghdl/ghdl-yosys-plugin">Plugin</a> integrates <a href="https://github.com/ghdl/ghdl">GHDL</a> into <a href="https://github.com/YosysHQ/yosys">Yosys</a> in order to allow VHDL synthesis.

### Requirements
We tested the following installation guide on an Ubuntu-22.04 subsystem (wsl 2) with the following requirements:

`GNAT 10.4.0` `Yosys 0.24+1` `GHDL 3.0.0-dev`

### Installation
Please follow the installation guides above to install GHDL any Yosys and follow the instructions below in order to install the <a href="https://github.com/ghdl/ghdl-yosys-plugin">Plugin</a>

1. `git clone https://github.com/ghdl/ghdl-yosys-plugin` to clone the git repository
2. `cd ghdl-yosys-plugin` to change to the "ghdl-yosys-plugin" directory
3. `make` to build the <a href="https://github.com/ghdl/ghdl-yosys-plugin">Plugin</a>
4. `make install` to install the <a href="https://github.com/ghdl/ghdl-yosys-plugin">Plugin</a>
