# Console Output
PROLEAD continuously displays the current evaluation results on the console. An exemplary console output (for the evaluation of first-order security in normal mode) is given below:

```
----------------------------------------------------------------------------------------------------------------------------------
| #Standard Probes | #Extended Probes | Security Order | Distance | #Entries in Report | #Probing Sets | Maximum #Probes per Set |
----------------------------------------------------------------------------------------------------------------------------------
|               21 |               25 |              1 |       10 |                 10 |            84 |                      18 |
----------------------------------------------------------------------------------------------------------------------------------

Evaluate security under the robust probing model!
----------------------------------------------------------------------------------------------------------------------------------------
|      Elapsed Time |      Required Ram |  Processed Simulations | Probing Set with highest Information Leakage |  -log10(p) |  Status |
----------------------------------------------------------------------------------------------------------------------------------------
|         9.747594s |        4.001968GB |        1536000 / 20325 |                                   y1_n14 (2) |   1.062213 |    OKAY |
----------------------------------------------------------------------------------------------------------------------------------------
``` 

The first table gives initial information about the following evaluation procedure. This table can be helpful to guess the required evaluation time and the amount of RAM before the evaluation. Hence, the user can save time as the settings can be adjusted before the evaluation has started. All parameters in the table, which are not given in `config.set` are described below:

| Output Parameter | Description |
| ---------------- | ----------- |
| `#Standard Probes` | Gives the total number of different standard probes. Standard probes can be placed on all register inputs and primary outputs. |
| `#Extended Probes` | Gives the total number of probe extensions. More precisely, `#Extended Probes` is the sum of all probes than are probed due to a probe extension procedure. |
| `#Probing Sets` | Gives the total number of different probing sets which are considered by PROLEAD. |
| `Maximum #Probes per Set` | Gives the maximum number of probe extensions of a single probing set. Note that the maximumnumber of probes per set is only given if `no_of_probe_steps` equals one. |

The second table continuously monitors the evaluation results. Hence, the table shows information regarding the progression of different parameters. All parameters are defined below:

| Output Parameter | Description |
| ---------------- | ----------- |
| `Elapsed Time` | Specifies how much time has been required for the evaluation. |
| `Required Ram` | Specifies how much RAM PROLEAD has used so far. |
| `Processed Simulations` | Specifies how many simulations PROLEAD has processed so far. If PROLEAD evaluates in the normal mode, the number of simulations, required to achieve a predefined statistical power, is given after a backslash. For the given example, this means that PROLEAD has processed 1536000 simulations while only 20325 simulations are necessary to achieve the desired statistical power.  |
| `Probing Set with highest Information Leakage` | Gives the probing set with the highest information leakage. We show all standard probes of the set by their signal name and the corresponding clock cycle in curly brackets. For the given example, this means that a probe placed on the wire with name `y1_n14` and recording during clock cycle 2 results in the highest information leakage. |
| `-log10(p)` | Gives the smallest result of the g-test in the logarithmic domain. Hence, we print the largest `-log10(p)`-value. For the given example, this means that the largest `-log10(p)`-value is 1.062213. |
| `Status` | Gives the interpretation of the `-log10(p)` value. Usually, `Status` equals `OKAY` if `-log10(p)` is smaller than 5. This means that PROLEAD has not detected any leakage. If `-log10(p)` is larger or equal than 5, we set `Status` to `LEAKAGE` indicating that PROLEAD has detect information leakage. |

# Reports
Moreover, PROLEAD creates detailed reports after each simulation step. An exemplary report (for the evaluation of first-order security in normal mode) is given below: 

```
Report file after 1536000 simulations:

1.) Summary of most leaking (and already active) probing sets per clock cycle: 

Cycle 2: @[y1_n14(2)] ==> [sboxIn2[0](2), sboxIn2[1](2), sboxIn2[2](2), en(2), g1Reg[3](2), sboxIn2[0](1), sboxIn2[1](1), sboxIn2[2](1), en(1), g1Reg[3](1)] -log10(p) = 1.06221 --> OKAY
Cycle 3: @[y1_n14(3)] ==> [sboxIn2[0](3), sboxIn2[1](3), sboxIn2[2](3), en(3), g1Reg[3](3), sboxIn2[0](2), sboxIn2[1](2), sboxIn2[2](2), en(2), g1Reg[3](2)] -log10(p) = 1.06221 --> OKAY
Cycle 4: @[y1_n14(4)] ==> [sboxIn2[0](4), sboxIn2[1](4), sboxIn2[2](4), en(4), g1Reg[3](4), sboxIn2[0](3), sboxIn2[1](3), sboxIn2[2](3), en(3), g1Reg[3](3)] -log10(p) = 1.06221 --> OKAY
Cycle 5: @[y1_n14(5)] ==> [sboxIn2[0](5), sboxIn2[1](5), sboxIn2[2](5), en(5), g1Reg[3](5), sboxIn2[0](4), sboxIn2[1](4), sboxIn2[2](4), en(4), g1Reg[3](4)] -log10(p) = 1.06221 --> OKAY

2.) Summary of the most leakging (and already active) probing sets: 

@[y1_n14(2)] ==> [sboxIn2[0](2), sboxIn2[1](2), sboxIn2[2](2), en(2), g1Reg[3](2), sboxIn2[0](1), sboxIn2[1](1), sboxIn2[2](1), en(1), g1Reg[3](1)] -log10(p) = 1.06221 --> OKAY
@[y1_n14(5)] ==> [sboxIn2[0](5), sboxIn2[1](5), sboxIn2[2](5), en(5), g1Reg[3](5), sboxIn2[0](4), sboxIn2[1](4), sboxIn2[2](4), en(4), g1Reg[3](4)] -log10(p) = 1.06221 --> OKAY
@[y1_n14(4)] ==> [sboxIn2[0](4), sboxIn2[1](4), sboxIn2[2](4), en(4), g1Reg[3](4), sboxIn2[0](3), sboxIn2[1](3), sboxIn2[2](3), en(3), g1Reg[3](3)] -log10(p) = 1.06221 --> OKAY
@[y1_n14(3)] ==> [sboxIn2[0](3), sboxIn2[1](3), sboxIn2[2](3), en(3), g1Reg[3](3), sboxIn2[0](2), sboxIn2[1](2), sboxIn2[2](2), en(2), g1Reg[3](2)] -log10(p) = 1.06221 --> OKAY
@[y1_n15(5)] ==> [sboxIn2[1](5), sboxIn2[2](5), en(5), g1Reg[2](5), sboxIn2[1](4), sboxIn2[2](4), en(4), g1Reg[2](4)] -log10(p) = 1.01901 --> OKAY
@[y1_n15(4)] ==> [sboxIn2[1](4), sboxIn2[2](4), en(4), g1Reg[2](4), sboxIn2[1](3), sboxIn2[2](3), en(3), g1Reg[2](3)] -log10(p) = 1.01901 --> OKAY
@[y1_n15(3)] ==> [sboxIn2[1](3), sboxIn2[2](3), en(3), g1Reg[2](3), sboxIn2[1](2), sboxIn2[2](2), en(2), g1Reg[2](2)] -log10(p) = 1.01901 --> OKAY
@[y1_n15(2)] ==> [sboxIn2[1](2), sboxIn2[2](2), en(2), g1Reg[2](2), sboxIn2[1](1), sboxIn2[2](1), en(1), g1Reg[2](1)] -log10(p) = 1.01901 --> OKAY
@[share3[0](2)] ==> [g1Reg[2](2), g1Reg[1](2), g1Reg[0](2), g2Reg[0](2), g2Reg[2](2), g1Reg[2](1), g1Reg[1](1), g1Reg[0](1), g2Reg[0](1), g2Reg[2](1)] -log10(p) = 0.784031 --> OKAY
@[share3[0](5)] ==> [g1Reg[2](5), g1Reg[1](5), g1Reg[0](5), g2Reg[0](5), g2Reg[2](5), g1Reg[2](4), g1Reg[1](4), g1Reg[0](4), g2Reg[0](4), g2Reg[2](4)] -log10(p) = 0.784031 --> OKAY
```

The report is divided into two sections. The first section shows the probing set resulting in the highest information leakage during a particular clock cycle. If a probing set contains probes recording during different clock cycles, the highest clock cycle counts. The second section shows the largest probing sets of the whole evaluation. Each probing set is given as follows:

```
@[standard probes] ==> [extended probes] -log10(p) = ... --> OKAY/LEAKAGE
```

# Interpretation
The decisive question at the end of the evaluation is whether the investigated design is safe or unsafe in the model under consideration. Since PROLEAD is not a formal verification tool, this question is not always easy to answer. In the following we want to discuss possible uncertainties.

## False Positives
Internally, we report a design as insecure if the `-log10(p)`-value of at least one probing set exceeds the threshold of 5.0, corresponding to a false-positive probability of 0.00001. However, the higher the number of considered probing sets is, the greater the chance that false positives will occur. Hence, exceeding the 5.0 threshold is not a strict criterion for insecure designs. If you evaluate a design and PROLEAD reports `LEAKAGE`, we recommend continuing the evaluation and considering the course of the `-log10(p)`-values. If the values do not grow in the further progression taking more simulations into account, the reported leakage probably occurred due to a false positive.

## False Negatives
PROLEAD does not consider all possible circuit states but simulates random inputs. Hence, if PROLEAD simulates too few inputs, leakage may remain undetected. In order to obtain the most reliable results, we specify the following rules for evaluations.

* Consider at least 100 million simulations for any hardware evaluation in the compact mode.
* Consider at least 1 million simulations for any software evaluation in the compact mode.
* Consider enough traces to detect any leakage with an effect size of 0.1 or larger for any evaluation in the normal mode.
