# Name
no_of_initial_clock_cycles

# Type
string

# Default
none

# Description
Specifies the input sequence to initialize the simulation. After the initialization, all manually defined signals hold their value until the simulation ends. During initialization, the user must specify all input signals for each clock cycle. For each simulation, PROLEAD randomly selects a group and shares the group value. To assign a share of the group value to a signal it can be assigned with the group_in signal while the following number denotes the share and the range of the group value. For example, group_in0[7:0] denotes the first share of the first eight bits of the group value. If the user does not specify no_of_initial_inputs only changing inputs must be defined. Here, the user must set the number of changing signals for each clock cycle individually before defining the signals. If the input signals should keep their value for several clock cycles this can be indicated by the * operator and the number of clock cycles to keep the state.

# Examples
```
% Example

no_of_initial_clock_cycles
2 % The initialization sequence takes two clock cycles.

% Assume that 'SharedInput' stores two 7-bit shares.
[7:0] SharedInput group_in0[7:0] % Set the first share depending on the defined groups.
[15:8] SharedInput group_in1[7:0] % Set the second share depending on the defined groups.
Reset 1'b1 % Set the reset signal to one.

% Initialize inputs to zero for the second clock cycle.
% PROLEAD keeps this input state for the rest of the simulation.
[7:0] SharedInput 8'h00
[15:8] SharedInput 8'h00
Reset  1'b0
```

```
% Example

no_of_initial_clock_cycles
2 % The initialization sequence takes two clock cycles.

3
[7:0] SharedInput group_in0[7:0] % Set first share.
[15:8] SharedInput group_in1[7:0] % Set second share.
Reset 1'b1 % Set the reset signal.

1
Reset  1'b0 % Change to reset signal to zero.
% Keep the value of both shares.
```

```
% Example

no_of_initial_clock_cycles
6 % The initialization sequence takes two clock cycles.

5*3 % Keep this state for 5 clock cycles
[7:0] SharedInput group_in0[7:0] % Set first share.
[15:8] SharedInput group_in1[7:0] % Set second share.
Reset 1'b1 % Set the reset signal.

1
Reset  1'b0 % Change to reset signal to zero.
% Keep the value of both shares.
```